/*
** dmadump.c - utility to display or write data using a DMA engine on an
**             Alpha Data Gen 3 reconfigurable computing device.
**
** (C) Copyright 2016 Alpha Data
**
** This utility functions in a similar way to the "dump2" utility, but uses DMA
** transfers (bus mastering) instead of Direct Slave (CPU-initiated) transfers.
**
** Note that each DMA engine has its own OCP address space, separate to the other
** DMA engines and to the Direct Slave OCP channel.
**
** Therefore, unless an FPGA design explicitly routes DMA OCP channels and Direct
** Slave channels to the same OCP endpoint (for example, the BlockRAM in the UBER
** FPGA design), data written using dump/dmadump will not be visible when
** attempting to read it back using dmadump/dump.
**
** This utility does not make any attempt to aggregate write data items into one
** large chunk before writing, so a write command that writes three 32-bit values
** will typically result in three DMA write transfers of four bytes each.
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   define _stscanf_s(buffer, ...) _stscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   define _tcscpy_s(x, y, z) tcscpy(x, z)
#   define _tcsnset_s(buffer, max, c, n) _tcsnset(buffer, c, n)
#endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tscanf_s scanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcslen(x) strlen(x)
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _totlower(c) tolower(c)
# define _gettc getc

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <assert.h>
#include <ctype.h>
#include <limits.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <num_token.h>
#include <platform_endian.h>

/* Exit codes */
#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_UNRECOGNIZED_COMMAND    (2) /* Invalid command passed as 1st positional argument */
#define EXIT_ALLOCATION_FAILURE      (3) /* A memory allocation failed */
#define EXIT_DEVICE_OPEN_ERROR       (4) /* Failed to open ADMXRC3 device */
#define EXIT_WRITEDMA_ERROR          (5) /* A DMA write call failed */
#define EXIT_READDMA_ERROR           (6) /* A DMA read call failed */
#define ERROR_COUNT_TOO_LARGE        (7) /* Requested byte count is too large */

#ifdef _WIN32
# define PROGRAM_NAME _T("dmadump.exe")
#else
# define PROGRAM_NAME _T("dmadump")
#endif

#if !defined(SIZE_MAX)
/* Hack for when SIZE_MAX missing; relies on two's complement representation */
# define SIZE_MAX ((size_t)(-1))
#endif

#if defined(_UNICODE)
# define EOF_VALUE WEOF
#else
# define EOF_VALUE EOF
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MIN(a, b) ((a) < (b) ? (a) : (b))

typedef enum _CommandType {
  CommandTypeFill,
  CommandTypeRead,
  CommandTypeWrite
} CommandType;

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

static void
dumpMemory(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int width,
  bool_t bBigEndian,
  bool_t bAlign,
  void* pBuffer,
  size_t n,
  uint64_t localAddress)
{
  static const unsigned int addressSize = sizeof(uint64_t);
  unsigned int addressStrLen = 3 + 2 * addressSize + (addressSize - 1) / 4;
  uint64_t limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint8_t* p8;
  TCHAR line[257];
  TCHAR* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten;
  size_t remaining;

  assert(indent < 128);
  assert(nColumn == 1 || nColumn == 2 || nColumn == 4 || nColumn == 8 || nColumn == 16 || nColumn == 32);
  assert(width == 1 || width == 2 || width == 4 || width == 8);

  rowBytes = nColumn * width;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  limit = localAddress + n;
  if (bAlign) {
    addressCol = localAddress & ~(uint64_t)columnMask;
    addressLo = ((unsigned int)localAddress) & columnMask;
  } else {
    addressCol = localAddress;
    addressLo = 0;
  }

  m = limit - localAddress;

  pPos = line;
  pPos[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent);
  pPos += MIN(indent, remaining);
  remaining -= MIN(indent, remaining);
  nwritten = _stprintf_s(pPos, remaining, _T("Dump of memory at 0x%08lX_%08lX + %llu(0x%llX) bytes:\n"),
    (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
    (unsigned long long)n, (unsigned long long)n);
  if (nwritten > 0) {
    pPos += nwritten;
    remaining -= nwritten;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 1 + addressStrLen);
  pPos += MIN(indent + 1 + addressStrLen, remaining);
  remaining -= MIN(indent + 1 + addressStrLen, remaining);
  for (i = 0; i < rowBytes; i += width) {
    switch (width) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lX "), (unsigned long)i);
      break;

    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lX "), (unsigned long)i);
      break;

    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lX "), (unsigned long)i);
      break;

    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lX "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
  }
  if (remaining) {
    *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
    *pPos++ = _T('\0');
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }

    memset(columnBuf, 0, rowBytes);
    memcpy(columnBuf + addressLo, (uint8_t*)pBuffer, chunk);
    pBuffer = (void*)((uint8_t*)pBuffer + chunk);

    p8 = (uint8_t*)columnBuf;

    pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
    pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
    nwritten = _stprintf_s(pPos, remaining, _T("0x%08lX_%08lX: "), (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += width) {
      switch (width) {
        case 0x1U:
          val8 = *p8++;
          if (i >= addressLo && i < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)val8);
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
          break;

        case 0x2U:
          if (bBigEndian) {
            val16 = (uint16_t)(((uint16_t)p8[0] << 8U) | ((uint16_t)p8[1] << 0U));
          } else {
            val16 = (uint16_t)(((uint16_t)p8[0] << 0U) | ((uint16_t)p8[1] << 8U));
          }
          p8 += 2;
          for (k = 2; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 3) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;

        case 0x4U:
          if (bBigEndian) {
            val32 = (uint32_t)(((uint32_t)p8[0] << 24U) | ((uint32_t)p8[1] << 16U) | ((uint32_t)p8[2] <<  8U) | ((uint32_t)p8[3] <<  0U));
          } else {
            val32 = (uint32_t)(((uint32_t)p8[0] <<  0U) | ((uint32_t)p8[1] <<  8U) | ((uint32_t)p8[2] << 16U) | ((uint32_t)p8[3] << 24U));
          }
          p8 += 4;
          for (k = 4; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
          }
          break;

        case 0x8U:
          if (bBigEndian) {
            val64  = (uint64_t)(((uint64_t)p8[0] << 56U) | ((uint64_t)p8[1] << 48U) | ((uint64_t)p8[2] << 40U) | ((uint64_t)p8[3] << 32U));
            val64 |= (uint64_t)(((uint64_t)p8[4] << 24U) | ((uint64_t)p8[5] << 16U) | ((uint64_t)p8[6] <<  8U) | ((uint64_t)p8[7] <<  0U));
          } else {
            val64  = (uint64_t)(((uint64_t)p8[0] <<  0U) | ((uint64_t)p8[1] <<  8U) | ((uint64_t)p8[2] << 16U) | ((uint64_t)p8[3] << 24U));
            val64 |= (uint64_t)(((uint64_t)p8[4] << 32U) | ((uint64_t)p8[5] << 40U) | ((uint64_t)p8[6] << 48U) | ((uint64_t)p8[7] << 56U));
          }
          p8 += 8;
          for (k = 8; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 4) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;
      }
      if (remaining) {
        *pPos++ = _T(' ');
        remaining--;
      }
    }

    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;

      if (remaining) {
        if (_istprint(c) && c != _T('\t')) {
          *pPos++ = c;
        } else {
          *pPos++ = _T('.');
        }
        remaining--;
      }
    }
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    localAddress += chunk;
  }
}

static int
fillMemoryUsingDma(
  unsigned int width,
  bool_t bBigEndian,
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  uint64_t localAddress,
  unsigned int count,
  uint64_t fillValue)
{
  ADMXRC3_STATUS status;
  uint8_t* pDataBlock = NULL;
  unsigned int pos;
  int ret = EXIT_OK; /* Assume success */
  
  pDataBlock = (uint8_t*)malloc(count);
  if (NULL == pDataBlock) {
    _tprintf(_T("*** Failed to allocate %u(0x%X) bytes for fill data buffer."), count, count);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  switch (width) {
  case 1:
    if (bBigEndian) {
      for (pos = 0; pos < count; pos++) {
        *(pDataBlock + pos) = cpu_to_be8((uint8_t)fillValue);
      }
    } else {
      for (pos = 0; pos < count; pos++) {
        *(pDataBlock + pos) = cpu_to_le8((uint8_t)fillValue);
      }
    }
    break;

  case 2:
    if (bBigEndian) {
      for (pos = 0; pos < count; pos += 2) {
        *((uint16_t*)(pDataBlock + pos)) = cpu_to_be16((uint16_t)fillValue);
      }
    } else {
      for (pos = 0; pos < count; pos += 2) {
        *((uint16_t*)(pDataBlock + pos)) = cpu_to_le16((uint16_t)fillValue);
      }
    }
    break;

  case 4:
    if (bBigEndian) {
      for (pos = 0; pos < count; pos += 4) {
        *((uint32_t*)(pDataBlock + pos)) = cpu_to_be32((uint32_t)fillValue);
      }
    } else {
      for (pos = 0; pos < count; pos += 4) {
        *((uint32_t*)(pDataBlock + pos)) = cpu_to_le32((uint32_t)fillValue);
      }
    }
    break;

  case 8:
    if (bBigEndian) {
      for (pos = 0; pos < count; pos += 8) {
        *((uint64_t*)(pDataBlock + pos)) = cpu_to_be64((uint64_t)fillValue);
      }
    } else {
      for (pos = 0; pos < count; pos += 8) {
        *((uint64_t*)(pDataBlock + pos)) = cpu_to_le64((uint64_t)fillValue);
      }
    }
    break;

  default:
    assert(FALSE);
  }

  status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, pDataBlock, count, localAddress);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to perform DMA write of %u(0x%X) bytes on DMA channel %u at local address 0x%08lX_%08lX: %s\n"),
      count, count,
      dmaChannel,
      (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
      ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_WRITEDMA_ERROR;
    goto done;
  }

done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  
  return ret;
}

static int
writeDataUsingDma(
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  uint64_t localAddress,
  void* pDataBlock,
  uint64_t count)
{
  ADMXRC3_STATUS status;
  
  if (count > SIZE_MAX) {
    _tprintf(_T("*** Byte count %llu(0x%llX) for DMA write is larger than SIZE_MAX."),
      (unsigned long long)count, (unsigned long long)count);
    return ERROR_COUNT_TOO_LARGE; /* Failure */
  }

  status = ADMXRC3_WriteDMAEx(
    hDevice,
    dmaChannel,
    0,
    pDataBlock,
    (size_t)count /* Cast safe due to above check */,
    localAddress);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to write %llu(0x%llX) bytes on DMA channel %u at local address 0x%08lX_%08lX: %s\n"),
      (unsigned long long)count, (unsigned long long)count,
      dmaChannel,
      (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
      ADMXRC3_GetStatusString(status, TRUE));
    return EXIT_WRITEDMA_ERROR; /* Failure */
  }
  
  return EXIT_OK; /* Success */
}

static int
dumpMemoryUsingDma(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int width,
  bool_t bBigEndian,
  bool_t bAlign,
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  unsigned int count,
  uint64_t localAddress)
{
  ADMXRC3_STATUS status;
  void* pDataBlock = NULL;
  int ret = EXIT_OK; /* Assume success */
  
  pDataBlock = malloc(count);
  if (NULL == pDataBlock) {
    _tprintf(_T("*** Failed to allocate %u(%X) bytes to hold read data\n"), count, count);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }
  
  status = ADMXRC3_ReadDMAEx(hDevice, dmaChannel, 0, pDataBlock, count, localAddress);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to read %u(0x%X) bytes on DMA channel %u from local address 0x%08lX_%08lX: %s\n"),
      count, count,
      dmaChannel,
      (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
      ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_READDMA_ERROR; /* Failure */
    goto done;
  }
  
  dumpMemory(indent, nColumn, width, bBigEndian, bAlign, pDataBlock, count, localAddress);
  
done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  
  return ret; /* Success */
}

#if defined(__VXWORKS__) || defined(__vxworks)  
 
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

#define FLAG_BYSERIAL   (0x1U) /* Interpret 'indexOrSerial' as a serial number */
#define FLAG_BIGENDIAN (0x20U) /* Read and write values in big-endian byte order */

static int
admxrc3DmaFill(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  uint64_t localAddress,
  unsigned int count,
  unsigned int width,
  uint64_t value)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  unsigned int alignment;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  alignment = count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  ret = fillMemoryUsingDma(width, bBigEndian, hDevice, dmaChannel, localAddress, count, value);
  if (ret) {
    goto done;
  }

done:
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  return ret;
}

int
admxrc3DmaDumpHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3DmaDumpHelp()\n"));
  _tprintf(_T("  int admxrc3DmaFB (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, val8)\n"));
  _tprintf(_T("  int admxrc3DmaXFB(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, val8)\n"));
  _tprintf(_T("  int admxrc3DmaFW (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, val16)\n"));
  _tprintf(_T("  int admxrc3DmaXFW(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, val16)\n"));
  _tprintf(_T("  int admxrc3DmaFD (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, val32)\n"));
  _tprintf(_T("  int admxrc3DmaXFD(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, val32)\n"));
  _tprintf(_T("  int admxrc3DmaFQ (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, val64(ULL))\n"));
  _tprintf(_T("  int admxrc3DmaXFQ(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, val64(ULL))\n"));
  _tprintf(_T("  int admxrc3DmaRB (indexOrSerial, flags, dmaChannel, localAddress,      byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaXRB(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaRW (indexOrSerial, flags, dmaChannel, localAddress,      byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaXRW(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaRD (indexOrSerial, flags, dmaChannel, localAddress,      byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaXRD(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaRQ (indexOrSerial, flags, dmaChannel, localAddress,      byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaXRQ(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3DmaWB (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaXWB(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaWW (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaXWW(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaWD (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaXWD(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3DmaWQ (indexOrSerial, flags, dmaChannel, localAddress,      byteCount, ...(ULL))\n"));
  _tprintf(_T("  int admxrc3DmaXWQ(indexOrSerial, flags, dmaChannel, localAddress(ULL), byteCount, ...(ULL))\n"));
  _tprintf(_T("'localAddress(ULL)' & 'val64(ULL)' are unsigned long long parameters.\n"));
  _tprintf(_T("'...(ULL)' is a variable number of unsigned long long parameters.\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL  0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_BIGENDIAN 0x%02X => Loads and stores use big-endian byte-order\n"), FLAG_BIGENDIAN);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3DmaXFB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  unsigned int val8)
{
  return admxrc3DmaFill(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 1, (uint64_t)val8);
}

int
admxrc3DmaFB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  unsigned int val8)
{
  return admxrc3DmaXFB(indexOrSerial, flags, dmaChannel, localAddress, count, val8);
}

int
admxrc3DmaXFW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  unsigned int val16)
{
  return admxrc3DmaFill(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 2, (uint64_t)val16);
}

int
admxrc3DmaFW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  unsigned int val16)
{
  return admxrc3DmaXFW(indexOrSerial, flags, dmaChannel, localAddress, count, val16);
}

int
admxrc3DmaXFD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  unsigned int val32)
{
  return admxrc3DmaFill(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 4, (uint64_t)val32);
}

int
admxrc3DmaFD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  unsigned int val32)
{
  return admxrc3DmaXFD(indexOrSerial, flags, dmaChannel, localAddress, count, val32);
}

int
admxrc3DmaXFQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  unsigned long long val64)
{
  return admxrc3DmaFill(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 8, (uint64_t)val64);
}

int
admxrc3DmaFQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  unsigned long long val64)
{
  return admxrc3DmaXFQ(indexOrSerial, flags, dmaChannel, localAddress, count, val64);
}

static int
admxrc3DmaDump(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  uint64_t localAddress,
  unsigned int count,
  unsigned int width)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  unsigned int alignment;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  alignment = (unsigned int)count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  ret = dumpMemoryUsingDma(0, 16 / width, width, bBigEndian, TRUE, hDevice, dmaChannel, count, localAddress);
  if (ret) {
    goto done;
  }

done:
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

int
admxrc3DmaXRB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count)
{
  return admxrc3DmaDump(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 1);
}

int
admxrc3DmaRB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count)
{
  return admxrc3DmaXRB(indexOrSerial, flags, dmaChannel, localAddress, count);
}

int
admxrc3DmaXRW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count)
{
  return admxrc3DmaDump(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 2);
}

int
admxrc3DmaRW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count)
{
  return admxrc3DmaXRW(indexOrSerial, flags, dmaChannel, localAddress, count);
}

int
admxrc3DmaXRD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count)
{
  return admxrc3DmaDump(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 4);
}

int
admxrc3DmaRD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count)
{
  return admxrc3DmaXRD(indexOrSerial, flags, dmaChannel, localAddress, count);
}

int
admxrc3DmaXRQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count)
{
  return admxrc3DmaDump(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 8);
}

int
admxrc3DmaRQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count)
{
  return admxrc3DmaXRQ(indexOrSerial, flags, dmaChannel, localAddress, count);
}

static int
dmaWriteArgList(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  uint64_t localAddress,
  uint64_t count,
  unsigned int width,
  va_list args)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  uint8_t* pDataBlock = NULL;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int alignment;
  size_t pos;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  if (count > SIZE_MAX) {
    _tprintf(_T("*** Cannot allocate %llu(0x%llX) bytes for fill data buffer; larger than SIZE_MAX."),
      (unsigned long long)count, (unsigned long long)count);
    ret = ERROR_COUNT_TOO_LARGE;
    goto done;
  }
  
  alignment = (unsigned int)count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %llu(0x%llX) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %llu(0x%llX)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  pDataBlock = (uint8_t*)malloc(count);
  if (NULL == pDataBlock) {
    _tprintf(_T("*** Failed to allocate %llu(0x%llX) bytes for fill data buffer."),
      (unsigned long long)count, (unsigned long long)count);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  pos = 0;
  while (pos < count) {
    switch (width) {
    case 1:
      val8 = (uint8_t)va_arg(args, unsigned int);
      *(pDataBlock + pos) = (uint8_t)(bBigEndian ? cpu_to_be8(val8) : cpu_to_le8(val8));
      pos += 1;
      break;
      
    case 2:
      val16 = (uint16_t)va_arg(args, unsigned int);
      *((uint16_t*)(pDataBlock + pos)) = (uint16_t)(bBigEndian ? cpu_to_be16(val16) : cpu_to_le16(val16));
      pos += 2;
      break;
      
    case 4:
      val32 = (uint32_t)va_arg(args, unsigned int);
      *((uint32_t*)(pDataBlock + pos)) = (uint32_t)(bBigEndian ? cpu_to_be32(val32) : cpu_to_le32(val32));
      pos += 4;
      break;
      
    case 8:
      val64 = (uint64_t)va_arg(args, unsigned long long);
      *((uint64_t*)(pDataBlock + pos)) = (uint64_t)(bBigEndian ? cpu_to_be64(val64) : cpu_to_le64(val64));
      pos += 8;
      break;
    }
  }
  
  ret = writeDataUsingDma(hDevice, dmaChannel, localAddress, pDataBlock, count);
  if (ret) {
    goto done;
  }

done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

int
admxrc3DmaXWB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 1, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaWB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 1, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaXWW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 2, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaWW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 2, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaXWD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 4, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaWD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 4, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaXWQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned long long localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 8, args);
  va_end(args);
  
  return ret;
}

int
admxrc3DmaWQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int dmaChannel,
  unsigned int localAddress,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = dmaWriteArgList(indexOrSerial, flags, dmaChannel, (uint64_t)localAddress, count, 8, args);
  va_end(args);
  
  return ret;
}
#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  unsigned int tokenFlags = 0;
  Option options[4];
  Arguments arguments = g_initArguments;
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  bool_t bBigEndian, bInterpretHex;
  CommandType command;
  unsigned int count, width, index, serial, dmaChannel;
  uint64_t localAddress;
  TCHAR* pCommand;
  uint8_t* pDataBlock = NULL;
  int ret = EXIT_OK;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("be");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("TRUE => use big-endian memory format");
  options[3].key = _T("hex");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("TRUE => interpret values as being hexadecimal");
  arguments.nOption = 4;
  arguments.option = options;
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("Command to perform: fb, fw, fd, fq, rb, rw, rd, rq, wb, ww, wd or wq");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("channel");
  paramMeaning[1] = _T("Index of DMA channel to use");
  paramType[2] = OptionUInt64;
  paramInfo[2] = _T("address");
  paramMeaning[2] = _T("Local / OCP address to fill, read or write");
  paramType[3] = OptionUInt;
  paramInfo[3] = _T("length");
  paramMeaning[3] = _T("Number of bytes to fill, read or write");
  paramType[4] = OptionString;
  paramInfo[4] = _T("item ...");
  paramMeaning[4] = _T("Fill value or data items to write");
  arguments.nParamType = 5;
  arguments.minNParam = 3;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;
  
  if (sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bBigEndian = options[2].value.booleanVal;
  bInterpretHex = options[3].value.booleanVal;
  if (bInterpretHex) {
    tokenFlags |= SAMPLE_TOKEN_FLAG_INTERPRET_HEX;
  }
  pCommand = arguments.param[0].pStringVal;
  dmaChannel = arguments.param[1].uintVal;
  localAddress = arguments.param[2].uint64Val;

  if (_tcsicmp(pCommand, _T("fb")) == 0) {
    width = 1;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fw")) == 0) {
    width = 2;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fd")) == 0) {
    width = 4;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fq")) == 0) {
    width = 8;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    command = CommandTypeWrite;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }

  if (arguments.nParam < 4) {
    count = width;
  } else {
    count = arguments.param[3].uintVal;
  }
  if (count % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n"), count, count, width);
    count &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%x)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  if (command == CommandTypeFill) {
    SampleTokenState state = g_sampleTokenStateInit;
    
    /* Attempt to parse 5th positional argument */
    if (arguments.nParam >= 5) {
      sampleTokenPumpString(&state, arguments.param[4].pStringVal, tokenFlags, NULL);
    }
    
    if (!sampleTokenParsed(&state)) {
      /* No 5th positional argument, so attempt to get token from stdin */
      _tprintf(_T("Enter fill value: "));
      fflush(stdout);
      state = g_sampleTokenStateInit;
      if (!feof(stdin)) {
        sampleTokenPumpFile(&state, stdin, tokenFlags);
      }
    }
    
    if (!sampleTokenParsed(&state)) {
      /* No token obtained, so do nothing */
      goto done;
    }
    
    ret = fillMemoryUsingDma(width, bBigEndian, hDevice, dmaChannel, localAddress, count, (uint64_t)state.result.value);
    if (ret) {
      goto done;
    }
  } else if (command == CommandTypeWrite) {
    SampleTokenState state;
    unsigned int pos;
    int i;

    pDataBlock = (uint8_t*)malloc(count);
    if (NULL == pDataBlock) {
      _tprintf(_T("*** Failed to allocate %u(%X) bytes to hold write values\n"), count, count);
      ret = EXIT_ALLOCATION_FAILURE;
      goto done;
    }
    
    i = 4;
    pos = 0;
    while (pos < count) {
      state = g_sampleTokenStateInit;
      
      /* Attempt to parse 5th and later positional arguments */
      if (i < arguments.nParam) {
        sampleTokenPumpString(&state, arguments.param[i].pStringVal, tokenFlags, NULL);
        i++;
        if (!sampleTokenParsed(&state)) {
          continue;
        }
        _tprintf(_T("0x%08lX_%08lX: 0x%llX\n"),
          (unsigned long)(uint32_t)((localAddress + pos) >> 32), (unsigned long)(uint32_t)(localAddress + pos),
          state.result.value);        
      }
      
      if (!sampleTokenParsed(&state)) {
        /* No positional arguments left, so attempt to get token from stdin */
        _tprintf(_T("0x%08lX_%08lX: "),
          (unsigned long)(uint32_t)((localAddress + pos) >> 32), (unsigned long)(uint32_t)(localAddress + pos));
        fflush(stdout);
        if (!feof(stdin)) {
          sampleTokenPumpFile(&state, stdin, tokenFlags);
        }          
      }
      
      if (!sampleTokenParsed(&state)) {
        /* No more tokens obtained, so quit from loop */
        break;
      }
      
      switch (width) {
      case 1:
        *(pDataBlock + pos) = bBigEndian ? cpu_to_be8((uint8_t)state.result.value) : cpu_to_le8((uint8_t)state.result.value);
        pos += 1;
        break;
        
      case 2:
        *((uint16_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be16((uint16_t)state.result.value) : cpu_to_le16((uint16_t)state.result.value);
        pos += 2;
        break;
        
      case 4:
        *((uint32_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be32((uint32_t)state.result.value) : cpu_to_le32((uint32_t)state.result.value);
        pos += 4;
        break;
        
      case 8:
        *((uint64_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be64((uint64_t)state.result.value) : cpu_to_le64((uint64_t)state.result.value);
        pos += 8;
        break;
      }
    }
    
    count = pos;
    ret = writeDataUsingDma(hDevice, dmaChannel, localAddress, pDataBlock, count);
    if (ret) {
      goto done;
    }
  } else if (command == CommandTypeRead) {
    ret = dumpMemoryUsingDma(0, 16 / width, width, bBigEndian, TRUE, hDevice, dmaChannel, count, localAddress);
    if (ret) {
      goto done;
    }
  } else {
    /* Should never get here */
    assert(FALSE);
  }

done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
