/*
** simple.c - Trivial example showing how to read and write registers
**            in the FPGA.
**
** (C) Copyright 2009-2011, 2016 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _tscanf_s(buffer, ...) _tscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
# define _gettc getc
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#if !defined(USE_POINTER_ACCESS)
# define USE_POINTER_ACCESS (1)
#endif

/* Exit codes */
#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_DEVICE_OPEN_ERROR       (2) /* Failed to open ADMXRC3 device */
#define EXIT_CARDINFO_ERROR          (3) /* Failed to get information about ADMXRC3 device */
#define EXIT_MAPWINDOW_ERROR         (4) /* Failed to map window into process' virtual address space */
#define EXIT_SETCLOCKFREQUENCY_ERROR (5) /* Failed to set clock generator frequency */
#define EXIT_CONFIGURATION_FAILED    (6) /* ADMXRC3_Configure{FromBuffer,FromFile} failed */
#define EXIT_BIT_FILENAME_FAILURE    (7) /* Failed to figure out full path of .bit file to use */

static int
doSimple(
  ADMXRC3_HANDLE hDevice,
  const TCHAR* pBitFilename)
{
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  volatile uint32_t* pFpgaSpace = NULL;
  unsigned long dataOut, dataIn;
  uint32_t val32;
  int err, ret = EXIT_OK;
  
  status = ADMXRC3_GetCardInfo(hDevice, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_CARDINFO_ERROR;
    goto done;
  }

#if USE_POINTER_ACCESS
  /* Get the address of FPGA space */
  status = ADMXRC3_MapWindow(hDevice, 0, 0x0, 0x10, (void*)&pFpgaSpace);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_MAPWINDOW_ERROR;
    goto done;
  }
#endif
    
  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hDevice, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1lf MHz: %s\n"), 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_SETCLOCKFREQUENCY_ERROR;
      goto done;
    }
  }

  if (NULL != pBitFilename) {
    status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pBitFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_CONFIGURATION_FAILED;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

  _tprintf(_T("Enter values to be nibble-reversed\n"));
#ifdef _WIN32
  _tprintf(_T("(CTRL-Z or enter 0 to exit)\n"));
#else
  _tprintf(_T("(CTRL-D or enter 0 to exit)\n"));
#endif

  do {
    err = _tscanf_s(_T("%lX"), &dataOut);
    if (0 == dataOut) {
      break;
    }
    
    if (err >= 1) {
      val32 = cpu_to_le32((uint32_t)dataOut);
#if USE_POINTER_ACCESS
      pFpgaSpace[0] = val32;
      dataIn = (unsigned long)pFpgaSpace[0];
#else
      ADMXRC3_Write(hDevice, 0, 0, 0, sizeof(uint32_t), &val32);
      ADMXRC3_Read(hDevice, 0, 0, 0, sizeof(uint32_t), &dataIn);
#endif
      dataIn = le32_to_cpu(dataIn);
      _tprintf(_T("OUT = 0x%08lx, IN = 0x%08lx\n"), (unsigned long)dataOut, (unsigned long)dataIn);      
    } else {
#ifdef _UNICODE
      wint_t c;
#else
      int c;
#endif
      
      if (0 == err) {
        /* Skip to next line */
        do {
          c = _gettc(stdin);
#ifdef _UNICODE
          if (WEOF == c) {
#else
          if (EOF == c) {
#endif
            goto done;
          }
        } while (c != _T('\n'));
      }
    }
  } while (EOF != err);

done:
  if (NULL != pFpgaSpace) {
    status = ADMXRC3_UnmapWindow(hDevice, (void*)pFpgaSpace);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("+++ Failed to unmap FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    }
  }

  return ret;
}

#if (defined(__VXWORKS__) || defined(__vxworks))
  
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/
 
/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL       (0x1U) /* Interpret 'index' as a serial number */
#define FLAG_USEUBER       (0x10U) /* Use UBER bitstream instead of SIMPLE */

int
admxrc3SimpleHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3SimpleHelp()\n"));
  _tprintf(_T("  int admxrc3Simple(indexOrSerial[, flags])\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL       0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_USEUBER        0x%02X => Use UBER bitstream instead of SIMPLE\n"), FLAG_USEUBER);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3Simple(
  unsigned int indexOrSerial, /* Index of device to open (OR serial number if 'flags' includes FLAG_BYSERIAL) */
  unsigned int flags)         /* See above flags */
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bUseUber = FALSE;
  TCHAR* pSdkBitDir;
  TCHAR* pBitFilename = NULL;
  bool_t bShouldConfigure;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial = TRUE;
  }
  if (flags & FLAG_USEUBER) {
    bUseUber = TRUE;
  }
  
  pSdkBitDir = sampleGetSdkBitDir();
  
  status = sampleOpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index/serial %lu: %s\n"), (unsigned long)bOpenBySerial, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  if (bUseUber) {
    pBitFilename = sampleConstructBitFilename(hDevice, 0, _T("uber"), _T("uber"), pSdkBitDir, &bShouldConfigure);
  } else {
    pBitFilename = sampleConstructBitFilename(hDevice, 0, _T("simple"), _T("simple"), pSdkBitDir, &bShouldConfigure);
  }
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doSimple(hDevice, bShouldConfigure ? pBitFilename : NULL);

done:
  if (NULL != pSdkBitDir) {
    free(pSdkBitDir);
  }
  if (NULL != pBitFilename) {
    free(pBitFilename);
  }
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR *argv[])
{
  ADMXRC3_STATUS status;
  unsigned int index, serial;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  Option options[3];
  Arguments arguments = g_initArguments;
  const TCHAR* pBitFilename = NULL;
  bool_t bShouldConfigure;
  bool_t bUseUber;
  int ret = EXIT_OK;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("uber");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("use uber design instead of simple");
  arguments.nOption = 3;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("simple"), argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bUseUber = options[2].value.booleanVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  if (bUseUber) {
    pBitFilename = sampleGetFpgaFilename(hDevice, 0, _T("uber"), _T("uber"), &arguments, &bShouldConfigure);
  } else {
    pBitFilename = sampleGetFpgaFilename(hDevice, 0, _T("simple"), _T("simple"), &arguments, &bShouldConfigure);
  }
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doSimple(hDevice, bShouldConfigure ? pBitFilename : NULL);

done:
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
