/*
** simpledma.c - Example that verifies the SIMPLEDMA example FPGA design.
**
** (C) Copyright 2012, 2016 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

#define EXIT_OK                       (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE         (1) /* Illegal basic command-line syntax */
#define EXIT_DEVICE_OPEN_ERROR        (2) /* Failed to open ADMXRC3 device */
#define EXIT_CARDINFO_ERROR           (3) /* Failed to get information about ADMXRC3 device */
#define EXIT_ALLOCATION_FAILURE       (4) /* A memory allocation failed */
#define EXIT_SETCLOCKFREQUENCY_ERROR  (5) /* Failed to set clock generator frequency */
#define EXIT_CONFIGURATION_FAILED     (6) /* ADMXRC3_Configure{FromBuffer,FromFile} failed */
#define EXIT_BIT_FILENAME_FAILURE     (7) /* Failed to figure out full path of .bit file to use */
#define EXIT_ILLEGAL_READ_LENGTH      (8) /* Illegal value for read length option */
#define EXIT_LOCK_ERROR               (9) /* Failed to lock DMA data buffer using ADMXRC3_Lock */
#define EXIT_READDMA_ERROR           (10) /* Failed to read data from FPGA using ADMXRC3_ReadDMA(Ex) */
#define EXIT_WRITEDMA_ERROR          (11) /* Failed to write data to FPGA using ADMXRC3_WriteDMA(Ex) */
#define EXIT_VERIFICATION_FAILED     (12) /* Data read from FPGA contained at least one incorrect byte */

/* Default maximum number of individual verification errors to report */
#define MAX_ERROR_REPORTED 20

/* Buffer size used for DMA transfers */
#define BUFFER_SIZE 0x100000U

/* Default number of bytes of data to read and display */
#define DEFAULT_READ_LENGTH (0x100U)

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

#if !defined(_MSC_VER) /* Non-MSVC - won't have secure strings library */
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

static void
dumpMemory(
	unsigned int indent,
	unsigned int nColumn,
	unsigned int radix,
  bool_t bBigEndian,
	bool_t bAlign,
  uint64_t offset,
	const void* pBuffer,
	size_t n)
{
  static const unsigned int ptrSize = sizeof(void*);
  unsigned int ptrStrLen = 3 + 2 * ptrSize + (ptrSize - 1) / 4;
	uint64_t address, limit, addressCol, m;
	unsigned int addressLo, columnMask;
	uint8_t* p;
	uint8_t* p8;
	TCHAR line[257];
	TCHAR* pPos;
	uint8_t columnBuf[256];
	uint8_t val8;
	uint16_t val16;
	uint32_t val32;
	uint64_t val64;
	unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten = 0;
  size_t remaining;

	assert(ptrSize == 4 || ptrSize == 8);

	if (indent > 128) {
		_tprintf(_T("*** dumpMemory: invalid value '%lu' for indent\n"), (unsigned long)indent);
		return;
	}
  /* SIZE_T_MAX should really be used here, but is not very portable */
  if (offset > (size_t)-1) {
		_tprintf(_T("*** dumpMemory: offset value '0x%llX' larger than SIZE_T_MAX\n"), (unsigned long long)offset);
		return;
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    _tprintf(_T("*** dumpMemory: invalid value '%lu' for nColumn\n"), (unsigned long)nColumn);
    return;
  }
	if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
		_tprintf(_T("*** dumpMemory: invalid value '%lu' for radix\n"), (unsigned long)radix);
		return;
	}

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
	columnMask = rowBytes - 1U;
	address = (uint64_t)(uintptr_t)pBuffer;
	limit = address + n;
	if (bAlign) {
		addressCol = offset & ~(uint64_t)columnMask;
		addressLo = ((unsigned int)address) & columnMask;
	} else {
		addressCol = offset;
		addressLo = 0;
	}

	m = limit - address;
	p = (uint8_t*)pBuffer;

	pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 1 + ptrStrLen);
  pPos += MIN(indent + 1 + ptrStrLen, remaining);
  remaining -= MIN(indent + 1 + ptrStrLen, remaining);
	for (i = 0; i < rowBytes; i += radix) {
		switch (radix) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lx "), (unsigned long)i);
      break;

    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lx "), (unsigned long)i);
      break;

    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lx "), (unsigned long)i);
      break;

    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lx "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
	}
  if (remaining) {
	  *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
	  *pPos++ = _T('\0');
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
	_tprintf(_T("%s"), line);

	while (m) {
		chunk = rowBytes - addressLo;
		if (chunk > m) {
			chunk = (unsigned int)m;
		}

		memset(columnBuf, 0, rowBytes);
		memcpy(columnBuf + addressLo, p, chunk);

		p8 = (uint8_t*)columnBuf;

		pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
		pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
		if (ptrSize == 4) {
			nwritten = _stprintf_s(pPos, remaining, _T("0x%08lx: "), (unsigned long)(uint32_t)addressCol);
		} else {
			nwritten = _stprintf_s(pPos, remaining, _T("0x%08lx_%08lx: "), (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
		}
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

		j = addressLo + chunk;
		for (i = 0; i < rowBytes; i += radix) {
			switch (radix) {
				case 0x1U:
					val8 = *p8++;
					if (i >= addressLo && i < j) {
						nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)val8);
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
					} else {
            if (remaining) {
  						*pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
  						*pPos++ = _T('.');
              remaining--;
            }
					}
					break;

				case 0x2U:
          if (bBigEndian) {
					  val16 = (uint16_t)(((uint16_t)p8[0] << 8U) | ((uint16_t)p8[1] << 0U));
          } else {
					  val16 = (uint16_t)(((uint16_t)p8[0] << 0U) | ((uint16_t)p8[1] << 8U));
          }
          p8 += 2;
					for (k = 2; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
						if (k == 3) {
              if (remaining) {
  							*pPos++ = _T('_');
                remaining--;
              }
						}
					}
					break;

				case 0x4U:
          if (bBigEndian) {
  					val32 = (uint32_t)(((uint32_t)p8[0] << 24U) | ((uint32_t)p8[1] << 16U) | ((uint32_t)p8[2] <<  8U) | ((uint32_t)p8[3] <<  0U));
          } else {
	  				val32 = (uint32_t)(((uint32_t)p8[0] <<  0U) | ((uint32_t)p8[1] <<  8U) | ((uint32_t)p8[2] << 16U) | ((uint32_t)p8[3] << 24U));
          }
          p8 += 4;
					for (k = 4; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
					}
					break;

				case 0x8U:
          if (bBigEndian) {
  					val64  = (uint64_t)(((uint64_t)p8[0] << 56U) | ((uint64_t)p8[1] << 48U) | ((uint64_t)p8[2] << 40U) | ((uint64_t)p8[3] << 32U));
	  				val64 |= (uint64_t)(((uint64_t)p8[4] << 24U) | ((uint64_t)p8[5] << 16U) | ((uint64_t)p8[6] <<  8U) | ((uint64_t)p8[7] <<  0U));
          } else {
  					val64  = (uint64_t)(((uint64_t)p8[0] <<  0U) | ((uint64_t)p8[1] <<  8U) | ((uint64_t)p8[2] << 16U) | ((uint64_t)p8[3] << 24U));
	  				val64 |= (uint64_t)(((uint64_t)p8[4] << 32U) | ((uint64_t)p8[5] << 40U) | ((uint64_t)p8[6] << 48U) | ((uint64_t)p8[7] << 56U));
          }
          p8 += 8;
					for (k = 8; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
						if (k == 4) {
              if (remaining) {
  							*pPos++ = _T('_');
                remaining--;
              }
						}
					}
					break;
			}
      if (remaining) {
			  *pPos++ = _T(' ');
        remaining--;
      }
		}

		p8 = (uint8_t*)columnBuf;
		for (i = 0; i < rowBytes; i++) {
			char c = (char)*p8++;

      if (remaining) {
			  if (_istprint(c) && c != _T('\t')) {
				  *pPos++ = c;
			  } else {
				  *pPos++ = _T('.');
			  }
        remaining--;
      }
		}
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
		addressLo = 0;
		addressCol += rowBytes;
		m -= chunk;
		p += chunk;
    offset += chunk;
	}
}

static int
doSimpleDma(
  ADMXRC3_HANDLE hDevice,
  const TCHAR* pBitFilename,
  uint64_t localAddress,
  unsigned int numByte,
  unsigned int maxError)
{
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  uint8_t* pCheckBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hCheckBuffer = 0;
  unsigned int numError = 0;
  unsigned int direction;
  unsigned int offset;
  uint32_t* p32;
  int ret = EXIT_OK;

  if (numByte > BUFFER_SIZE) {
    _tprintf(_T("*** Value for option '-n' must be no greater than %u\n"), BUFFER_SIZE);
    ret = EXIT_ILLEGAL_READ_LENGTH;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hDevice, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_CARDINFO_ERROR;
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hDevice, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1lf MHz: %s\n"), 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_SETCLOCKFREQUENCY_ERROR;
      goto done;
    }
  }

  pCheckBuffer = (uint8_t*)malloc(BUFFER_SIZE);
  if (NULL == pCheckBuffer) {
    _tprintf(_T("*** Failed to allocate check buffer of 0x%x bytes.\n"), (unsigned int)BUFFER_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  status = ADMXRC3_Lock(hDevice, pCheckBuffer, BUFFER_SIZE, &hCheckBuffer);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to lock check buffer for DMA transfers: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_LOCK_ERROR;
    goto done;
  }

  if (NULL != pBitFilename) {
    status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pBitFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_CONFIGURATION_FAILED;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

  _tprintf(_T("Reading using DMA channel 0 at OCP/AXI address 0x%llx...\n"), (unsigned long long)localAddress);
  status = ADMXRC3_ReadDMAEx(hDevice, 0, 0, pCheckBuffer, numByte, localAddress);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to read using DMA channel 0: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_READDMA_ERROR;
    goto done;
  }
  _tprintf(_T("Dump of data read from OCP/AXI address 0x%llx + 0x%x B:\n"), (unsigned long long)localAddress, numByte);
  dumpMemory(0, 4, 4, FALSE, FALSE, localAddress, pCheckBuffer, numByte);

  /* Check the data */
  p32 = (uint32_t*)pCheckBuffer;
  numError = 0U;
  for (offset = 0U; offset < numByte; offset += 4) {
    uint32_t exp32, val32;

    exp32 = offset >> 2;
    val32 = *p32++;
    if (val32 != exp32) {
      if (numError < maxError) {
        _tprintf(_T("*** Error at offset 0x%08x: expected=0x%08lx actual=0x%08lx\n"),
          offset, (unsigned long)exp32, (unsigned long)val32);
      } else if (numError == MAX_ERROR_REPORTED) {
        _tprintf(_T("*** More errors...\n"));
      }
      numError++;
    }
  }
  if (numError) {
    _tprintf(_T("*** FAILED with %llu (0x%llx) verification error(s)\n"), (unsigned long long)numError, (unsigned long long)numError);
    ret = EXIT_VERIFICATION_FAILED;
    goto done;
  }

  _tprintf(_T("Measuring throughput...                 \n"));
  for (direction = 0; direction < 2; direction++) {
    admxrc3_time_t start, finish, now;
    uint64_t total = 0;
    double elapsed, rate;

    start = timeGet();
    finish = timeAdd(start, timeMake(2, 0, FALSE));
    do {
      if (direction == 0) {
        status = ADMXRC3_WriteDMALockedEx(hDevice, 0, 0, hCheckBuffer, 0, BUFFER_SIZE, 0);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_WRITEDMA_ERROR;
          goto done;
        }
      } else {
        status = ADMXRC3_ReadDMALockedEx(hDevice, 0, 0, hCheckBuffer, 0, BUFFER_SIZE, 0);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_READDMA_ERROR;
          goto done;
        }
      }
      total += BUFFER_SIZE;
      now = timeGet();
    } while (timeBefore(now, finish));
    elapsed = timeToDouble(timeSubtract(now, start));
    rate = (double)total / elapsed;
    _tprintf(_T("Throughput %s is %.1f MiB/s\n"), (direction == 0) ? _T("from host to FPGA") : _T("from FPGA to host"), rate / 1048576.0);
  }

  _tprintf(_T("PASSED                                  \n"));

done:
  if (0 != hCheckBuffer) {
    ADMXRC3_Unlock(hDevice, hCheckBuffer);
  }
  if (NULL != pCheckBuffer) {
    free(pCheckBuffer);
  }
  
  return ret;
}

#if (defined(__VXWORKS__) || defined(__vxworks))
  
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/
 
/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL       (0x1U) /* Interpret 'index' as a serial number */

int
admxrc3SimpleDmaHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3SimpleDmaHelp()\n"));
  _tprintf(_T("  int admxrc3SimpleDma(indexOrSerial[, flags[, localAddressLo[, localAddressHi[, numByte[, maxError]]]]])\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL       0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3SimpleDma(
  unsigned int indexOrSerial,  /* Index of device to open (OR serial number if 'flags' includes FLAG_BYSERIAL) */
  unsigned int flags,          /* See above flags */
  uint32_t     localAddressLo, /* Low 32 bits of local (OCP/AXI) address at which to being reading */
  uint32_t     localAddressHi, /* High 32 bits of local (OCP/AXI) address at which to being reading */
  unsigned int numByte,        /* Number of bytes to read, or 0 for default */
  unsigned int maxError)       /* Maximum number of errors to display, or 0 for default */
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  uint64_t localAddress = (uint64_t)localAddressLo | ((uint64_t)localAddressHi << 32);
  bool_t bOpenBySerial = FALSE;
  TCHAR* pSdkBitDir = NULL;
  TCHAR* pBitFilename = NULL;
  bool_t bShouldConfigure;
  int ret = EXIT_OK;
  
  if (0 == numByte) {
    numByte = DEFAULT_READ_LENGTH;
  }
  if (0 == maxError) {
    maxError = MAX_ERROR_REPORTED;
  }
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial = TRUE;
  }
  
  pSdkBitDir = sampleGetSdkBitDir();
  
  status = sampleOpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index/serial %lu: %s\n"), (unsigned long)bOpenBySerial, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  pBitFilename = sampleConstructBitFilename(hDevice, 0, _T("simpledma"), _T("simpledma"), pSdkBitDir, &bShouldConfigure);
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doSimpleDma(hDevice, bShouldConfigure ? pBitFilename : NULL, localAddress, numByte, maxError);

done:
  if (NULL != pSdkBitDir) {
    free(pSdkBitDir);
  }
  if (NULL != pBitFilename) {
    free(pBitFilename);
  }
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}
  
#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR *argv[])
{
  ADMXRC3_STATUS status;
  unsigned int index, serial;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  Option options[5];
  Arguments arguments = g_initArguments;
  const TCHAR* pBitFilename = NULL;
  bool_t bShouldConfigure;
  unsigned int maxError;
  uint64_t localAddress;
  unsigned int numByte;
  int ret = EXIT_OK;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("address");
  options[2].type = OptionHex64;
  options[2].def.hex64Val = (uint64_t)0x0ULL;
  options[2].meaning = _T("OCP/AXI address from which to read (hexadecimal)");
  options[3].key = _T("maxerror");
  options[3].type = OptionUInt;
  options[3].def.uintVal = MAX_ERROR_REPORTED;
  options[3].meaning = _T("Maximum number of errors to display");
  options[4].key = _T("n");
  options[4].type = OptionUInt;
  options[4].def.uintVal = DEFAULT_READ_LENGTH;
  options[4].meaning = _T("Number of bytes to read");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("simpledma"), argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  localAddress = options[2].value.hex64Val;
  maxError = options[3].value.uintVal;
  numByte = options[4].value.uintVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  pBitFilename = sampleGetFpgaFilename(hDevice, 0, _T("simpledma"), _T("simpledma"), &arguments, &bShouldConfigure);
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doSimpleDma(hDevice, bShouldConfigure ? pBitFilename : NULL, localAddress, numByte, maxError);

done:
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
