/*
** File: model_admpcie8k5.c
** Project: ADMXRC3 module driver
** Purpose: ADM-PCIE-8K5 specific code, including the function to bootstrap
**          the device context.
**
** (C) Copyright Alpha Data 2016
*/

#include <df.h>

#include "device.h"
#include "info.h"
#include "model_boot.h"
#include <core/admpcie8k5.h>

static ADMXRC3_STATUS
getBankInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int bankIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETBANKINFO* pIoctl)
{
  VpdAdmpcie8k5* pVpd = (VpdAdmpcie8k5*)pVpdBuffer;
  uint8_t order;

  if (bankIndex >= DF_ARRAY_LENGTH(pVpd->rev0.sdram)) {
    return ADMXRC3_INVALID_INDEX;
  }
  order = pVpd->rev0.sdram[bankIndex].order;
  pIoctl->out.typeMask = ADMXRC3_BANK_SDRAM_DDR4;
  pIoctl->out.physicalDataWidth = 64U;
  pIoctl->out.physicalEccWidth = 8U;
  pIoctl->out.maximumFrequency = (uint64_t)pVpd->rev0.sdram[bankIndex].speed * 100000;
  pIoctl->out.minimumFrequency = 625000000U;
  pIoctl->out.connectivityMask = 0x1U;
  pIoctl->out.physicalSize = (uint64_t)0x1U << order;
  pIoctl->out.bPresent = (uint8_t)(order != 0);

  return ADMXRC3_SUCCESS;
}

static void
getCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl)
{
  VpdAdmpcie8k5* pVpd = (VpdAdmpcie8k5*)pVpdBuffer;
  unsigned int bankIndex;

  pIoctl->out.serialNumber = pVpd->rev0.serial;
  pIoctl->out.model = ADMXRC3_MODEL_ADMPCIE8K5;
  pIoctl->out.numMemoryBank = DF_ARRAY_LENGTH(pVpd->rev0.sdram);
  pIoctl->out.memoryBankPresent = 0;
  for (bankIndex = 0; bankIndex < DF_ARRAY_LENGTH(pVpd->rev0.sdram); bankIndex++) {
    if (0 != pVpd->rev0.sdram[bankIndex].order) {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << bankIndex);
    }
  }
}

static ADMXRC3_STATUS
getFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl)
{
  VpdAdmpcie8k5* pVpd = (VpdAdmpcie8k5*)pVpdBuffer;
  ADMXRC3_PACKAGE_TYPE package;
  ADMXRC3_FPGA_TYPE fpgaType;
  char speed[8];
  char stepping[8];

  switch (targetIndex) {
  case 0:
    fpgaType = (ADMXRC3_FPGA_TYPE)pVpd->rev0.fpga.device;
    break;

  default:
    return ADMXRC3_INVALID_INDEX;
  }

  package = ADMXRC3_PACKAGE_MAKEFLVA(1517);

  dfAssert(sizeof(char) == sizeof(uint8_t));

  dfStrCpy(speed, DF_ARRAY_LENGTH(speed), (const char*)pVpd->rev0.fpga.speedGrade);
  switch (pVpd->rev0.fpga.tempGrade) {
  case ADMPCIE8K5_VPD_TEMPGRADE_COMMERCIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "C");
    break;

  case ADMPCIE8K5_VPD_TEMPGRADE_INDUSTRIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "I");
    break;

  case ADMPCIE8K5_VPD_TEMPGRADE_EXTENDED:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "E");
    break;

  default:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "?");
    break;
  }

  dfStrCpy(stepping, DF_ARRAY_LENGTH(stepping), (const char*)pVpd->rev0.fpga.stepping);
  pIoctl->out.flags = ADMXRC3_FPGA_SPEEDVALID | ADMXRC3_FPGA_NOTCONFIGURABLE;
  if (stepping[0] != '\0') {
    pIoctl->out.flags |= ADMXRC3_FPGA_STEPPINGVALID;
  }
  setupIoctlsGetFpgaInfo(pDevCtx, pIoctl, fpgaType, package, speed, stepping, TRUE);

  return ADMXRC3_SUCCESS;
}

void
admxrc3BootstrapAdmpcie8k5(
	Admxrc3DeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("admxrc3BootstrapAdmpcie8k5: entered\n"));

  pDevCtx->methods.pGetBankInfo = getBankInfo;
  pDevCtx->methods.pGetCardInfo = getCardInfo;
  pDevCtx->methods.pGetFpgaInfo = getFpgaInfo;
  pDevCtx->bReverseSelectMapData = FALSE;
  pDevCtx->localAddressLimit = (uint64_t)(0x1ULL << 39U);
  pDevCtx->localWidth = 32U;
}
